#include <iostream>
#include "Adaptor.h"

#include <vtkCellData.h>
#include <vtkCellType.h>
#include <vtkCPDataDescription.h>
#include <vtkCPInputDataDescription.h>
#include <vtkCPProcessor.h>
#include <vtkCPPythonScriptPipeline.h>
#include <vtkDoubleArray.h>
#include <vtkFloatArray.h>
#include <vtkNew.h>
#include <vtkPoints.h>
#include <vtkPointData.h>
#include <vtkImageData.h>

namespace
{
	vtkCPProcessor* Processor = NULL;
	vtkImageData* VTKGrid;

	void BuildVTKGrid(int w, int h, int d)
	{
		if(VTKGrid == NULL) {
			VTKGrid = vtkImageData::New();
			int extent[6] = {0,w-1,0,h-1,0,d-1};
			VTKGrid->SetExtent(extent);
			VTKGrid->SetSpacing(1.0,1.0,1.0);
		}
	}

	void UpdateVTKAttributes(int w, int h, int d, double* field)
	{
		if(VTKGrid->GetCellData()->GetNumberOfArrays() == 0)
		{
			vtkNew<vtkDoubleArray> array;
			array->SetName("space");
			array->SetNumberOfComponents(1);
			VTKGrid->GetCellData()->AddArray(array.GetPointer());
		}

		vtkDoubleArray* space = vtkDoubleArray::SafeDownCast(
				VTKGrid->GetCellData()->GetArray("space"));
		// The pressure array is a scalar array so we can reuse
		// memory as long as we ordered the points properly.
		space->SetArray(field, static_cast<vtkIdType>(w*h*d), 1);
	}

	void BuildVTKDataStructures(int w, int h, int d, double* field)
	{
		if(VTKGrid == NULL)
		{
			// The grid structure isn't changing so we only build it
			// the first time it's needed. If we needed the memory
			// we could delete it and rebuild as necessary.
			VTKGrid = vtkImageData::New();
			BuildVTKGrid(w, h, d);
		}
		UpdateVTKAttributes(w, h, d, field);
	}
}

namespace Adaptor
{
	void Initialize(const std::string& fileName)
	{
		if(Processor == NULL)
		{
			Processor = vtkCPProcessor::New();
			Processor->Initialize();
		}
		else
		{
			Processor->RemoveAllPipelines();
		}

		vtkNew<vtkCPPythonScriptPipeline> pipeline;
		pipeline->Initialize(fileName.c_str());
		Processor->AddPipeline(pipeline.GetPointer());
	}

	void Finalize()
	{
		if(Processor)
		{
			Processor->Delete();
			Processor = NULL;
		}
		if(VTKGrid)
		{
			VTKGrid->Delete();
			VTKGrid = NULL;
		}
	}

	void CoProcess(int iteration, int w, int h, int d, double* field)
	{
		vtkNew<vtkCPDataDescription> dataDescription;
		dataDescription->AddInput("input");
		dataDescription->SetTimeData(0.0,iteration);
		
		if(Processor->RequestDataDescription(dataDescription.GetPointer()) != 0)
		{
			BuildVTKDataStructures(w,h,d,field);
			dataDescription->GetInputDescriptionByName("input")->SetGrid(VTKGrid);
			Processor->CoProcess(dataDescription.GetPointer());
		}
	}
} // end of Catalyst namespace
